import pytest
from pydantic import BaseModel, ValidationError

from crypticorn.common import (
    FilterParams,
    HeavyPageSortFilterParams,
    HeavyPaginationParams,
    PageFilterParams,
    PageSortFilterParams,
    PageSortParams,
    PaginationParams,
    SortFilterParams,
    SortParams,
)


class Item(BaseModel):
    name: str
    value: int


@pytest.mark.asyncio
async def test_pagination_params():
    # Test default values
    params = PaginationParams[Item]()
    assert params.page == 1
    assert params.page_size == 10

    # Test custom values
    params = PaginationParams[Item](page=2, page_size=20)
    assert params.page == 2
    assert params.page_size == 20

    # Test page_size validation (should be between 1 and 100)
    with pytest.raises(ValidationError):
        PaginationParams[Item](page_size=0)

    with pytest.raises(ValidationError):
        PaginationParams[Item](page_size=101)


@pytest.mark.asyncio
async def test_heavy_pagination_params():
    # Test default values
    params = HeavyPaginationParams[Item]()
    assert params.page == 1
    assert params.page_size == 100

    # Test custom values
    params = HeavyPaginationParams[Item](page=2, page_size=200)
    assert params.page == 2
    assert params.page_size == 200

    # Test page_size validation (should be between 1 and 1000)
    with pytest.raises(ValidationError):
        HeavyPaginationParams[Item](page_size=0)

    with pytest.raises(ValidationError):
        HeavyPaginationParams[Item](page_size=1001)


@pytest.mark.asyncio
async def test_sort_params():
    # Test that SortParams requires BaseModel generic
    with pytest.raises(
        TypeError,
        match="SortParams must be used with a Pydantic BaseModel as a generic parameter",
    ):
        SortParams[int]()

    # Test invalid sort field
    with pytest.raises(
        ValueError,
        match="Invalid field: 'foo'. Must be one of: \\['name', 'value'\\]",
    ):
        SortParams[Item](sort_by="foo", sort_order="asc")

    # Test that sort_by and sort_order must be provided together
    with pytest.raises(
        ValueError, match="sort_order and sort_by must be provided together"
    ):
        SortParams[Item](sort_by="name")

    with pytest.raises(
        ValueError, match="sort_order and sort_by must be provided together"
    ):
        SortParams[Item](sort_order="asc")

    # Test valid combination
    params = SortParams[Item](sort_by="name", sort_order="asc")
    assert params.sort_by == "name"
    assert params.sort_order == "asc"

    # Test default values
    params = SortParams[Item]()
    assert params.sort_by is None
    assert params.sort_order is None


@pytest.mark.asyncio
async def test_sort_order_validation():
    # Test invalid order values - this should be caught by Pydantic's Literal validation
    with pytest.raises(ValidationError):
        SortParams[Item](sort_by="name", sort_order="invalid")

    # Test valid order values
    params = SortParams[Item](sort_by="name", sort_order="asc")
    assert params.sort_order == "asc"

    params = SortParams[Item](sort_by="name", sort_order="desc")
    assert params.sort_order == "desc"


@pytest.mark.asyncio
async def test_filter_params():
    # Test that FilterParams requires BaseModel generic
    with pytest.raises(
        TypeError,
        match="FilterParams must be used with a Pydantic BaseModel as a generic parameter",
    ):
        FilterParams[int](filter_by="name", filter_value="test")

    # Test that filter_value must be provided when filter_by is set
    with pytest.raises(
        ValueError, match="filter_by and filter_value must be provided together"
    ):
        FilterParams[Item](filter_by="name")

    # Test invalid filter field
    with pytest.raises(
        ValueError,
        match="Invalid field: 'foo'. Must be one of: \\['name', 'value'\\]",
    ):
        FilterParams[Item](filter_by="foo", filter_value="test")

    # Test valid filter
    params = FilterParams[Item](filter_by="name", filter_value="test")
    assert params.filter_by == "name"
    assert params.filter_value == "test"

    # Test default values
    params = FilterParams[Item]()
    assert params.filter_by is None
    assert params.filter_value is None


@pytest.mark.asyncio
async def test_sort_filter_params():
    # Test combined functionality
    params = SortFilterParams[Item](
        sort_by="value",
        sort_order="desc",
        filter_by="name",
        filter_value="test",
    )
    assert params.sort_by == "value"
    assert params.sort_order == "desc"
    assert params.filter_by == "name"
    assert params.filter_value == "test"

    # Test default values
    params = SortFilterParams[Item]()
    assert params.sort_by is None
    assert params.sort_order is None
    assert params.filter_by is None
    assert params.filter_value is None

    # Test sort validation still works
    with pytest.raises(
        ValueError, match="sort_order and sort_by must be provided together"
    ):
        SortFilterParams[Item](sort_by="name")

    # Test filter validation still works
    with pytest.raises(
        ValueError, match="filter_by and filter_value must be provided together"
    ):
        SortFilterParams[Item](filter_by="name")


@pytest.mark.asyncio
async def test_page_filter_params():
    # Test combined functionality
    params = PageFilterParams[Item](
        page=2,
        page_size=20,
        filter_by="name",
        filter_value="test",
    )
    assert params.page == 2
    assert params.page_size == 20
    assert params.filter_by == "name"
    assert params.filter_value == "test"

    # Test default values
    params = PageFilterParams[Item]()
    assert params.page == 1
    assert params.page_size == 10
    assert params.filter_by is None
    assert params.filter_value is None

    # Test filter validation still works
    with pytest.raises(
        ValueError, match="filter_by and filter_value must be provided together"
    ):
        PageFilterParams[Item](filter_by="name")


@pytest.mark.asyncio
async def test_page_sort_params():
    # Test combined functionality
    params = PageSortParams[Item](
        page=2,
        page_size=20,
        sort_by="value",
        sort_order="desc",
    )
    assert params.page == 2
    assert params.page_size == 20
    assert params.sort_by == "value"
    assert params.sort_order == "desc"

    # Test default values
    params = PageSortParams[Item]()
    assert params.page == 1
    assert params.page_size == 10
    assert params.sort_by is None
    assert params.sort_order is None

    # Test sort validation still works
    with pytest.raises(
        ValueError, match="sort_order and sort_by must be provided together"
    ):
        PageSortParams[Item](sort_by="name")


@pytest.mark.asyncio
async def test_page_sort_filter_params():
    # Test combined functionality
    params = PageSortFilterParams[Item](
        page=2,
        page_size=20,
        sort_by="value",
        sort_order="desc",
        filter_by="name",
        filter_value="test",
    )
    assert params.page == 2
    assert params.page_size == 20
    assert params.sort_by == "value"
    assert params.sort_order == "desc"
    assert params.filter_by == "name"
    assert params.filter_value == "test"

    # Test default values
    params = PageSortFilterParams[Item]()
    assert params.page == 1
    assert params.page_size == 10
    assert params.sort_by is None
    assert params.sort_order is None
    assert params.filter_by is None
    assert params.filter_value is None

    # Test sort validation still works
    with pytest.raises(
        ValueError, match="sort_order and sort_by must be provided together"
    ):
        PageSortFilterParams[Item](sort_by="name")

    # Test filter validation still works
    with pytest.raises(
        ValueError, match="filter_by and filter_value must be provided together"
    ):
        PageSortFilterParams[Item](filter_by="name")


@pytest.mark.asyncio
async def test_heavy_page_sort_filter_params():
    # Test combined functionality
    params = HeavyPageSortFilterParams[Item](
        page=2,
        page_size=200,
        sort_by="value",
        sort_order="desc",
        filter_by="name",
        filter_value="test",
    )
    assert params.page == 2
    assert params.page_size == 200
    assert params.sort_by == "value"
    assert params.sort_order == "desc"
    assert params.filter_by == "name"
    assert params.filter_value == "test"

    # Test default values
    params = HeavyPageSortFilterParams[Item]()
    assert params.page == 1
    assert params.page_size == 100
    assert params.sort_by is None
    assert params.sort_order is None
    assert params.filter_by is None
    assert params.filter_value is None

    # Test page_size validation (should be between 1 and 1000)
    with pytest.raises(ValidationError):
        HeavyPageSortFilterParams[Item](page_size=0)

    with pytest.raises(ValidationError):
        HeavyPageSortFilterParams[Item](page_size=1001)


@pytest.mark.asyncio
async def test_field_type_validation():
    # Test that invalid type for filter_value raises error
    with pytest.raises(
        ValueError, match="Expected <class 'int'> for field value, got <class 'str'>"
    ):
        PageSortFilterParams[Item](
            filter_by="value", filter_value="not_a_number"
        )  # tries to coerce to int, but fails

    # Test that valid type works
    params = PageSortFilterParams[Item](
        filter_by="value", filter_value="42"
    )  # in fact it should be an int but it comes as a string anyways from the request. the type is enforced by the model validator
    assert params.filter_value == 42

    # Test type coercion
    params = PageSortFilterParams[Item](filter_by="name", filter_value="1")
    assert (
        params.filter_value == "1"
    )  # since name is a str, it will be coerced to a string

    params = PageSortFilterParams[Item](filter_by="name", filter_value="test")
    assert params.filter_value == "test"



# coding: utf-8

"""
Marketcap Service API

API for retrieving historical marketcap data, available exchanges, and indicators.

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, ClassVar, Dict, List, Optional, Set, Union

from pydantic import BaseModel, ConfigDict, Field, StrictFloat, StrictInt
from typing_extensions import Self


class OHLCV(BaseModel):
    """
    Open, High, Low, Close, Volume candlestick data with optional market cap information.
    """  # noqa: E501

    timestamp: StrictInt = Field(description="Unix timestamp of the candle period")
    open: Union[StrictFloat, StrictInt] = Field(
        description="Opening price of the period"
    )
    high: Union[StrictFloat, StrictInt] = Field(
        description="Highest price during the period"
    )
    low: Union[StrictFloat, StrictInt] = Field(
        description="Lowest price during the period"
    )
    close: Union[StrictFloat, StrictInt] = Field(
        description="Closing price of the period"
    )
    volume: Union[StrictFloat, StrictInt] = Field(
        description="Trading volume during the period"
    )
    marketcap: Optional[Union[StrictFloat, StrictInt]] = None
    __properties: ClassVar[List[str]] = [
        "timestamp",
        "open",
        "high",
        "low",
        "close",
        "volume",
        "marketcap",
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of OHLCV from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # set to None if marketcap (nullable) is None
        # and model_fields_set contains the field
        if self.marketcap is None and "marketcap" in self.model_fields_set:
            _dict["marketcap"] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of OHLCV from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "timestamp": obj.get("timestamp"),
                "open": obj.get("open"),
                "high": obj.get("high"),
                "low": obj.get("low"),
                "close": obj.get("close"),
                "volume": obj.get("volume"),
                "marketcap": obj.get("marketcap"),
            }
        )
        return _obj



"""General utility functions and helper methods used across the codebase."""

import random
import string
import warnings
from datetime import datetime
from decimal import Decimal
from typing import Any, Union

import typing_extensions

from crypticorn.common.exceptions import ApiError, ExceptionContent, HTTPException
from crypticorn.common.warnings import CrypticornDeprecatedSince25


def throw_if_none(
    value: Any,
    message: str = "Object not found",
) -> None:
    """Throws an FastAPI HTTPException if the value is None. https://docs.python.org/3/library/stdtypes.html#truth-value-testing"""
    if value is None:
        raise HTTPException(
            content=ExceptionContent(error=ApiError.OBJECT_NOT_FOUND, message=message)
        )


def throw_if_falsy(
    value: Any,
    message: str = "Object not found",
) -> None:
    """Throws an FastAPI HTTPException if the value is False. https://docs.python.org/3/library/stdtypes.html#truth-value-testing"""
    if not value:
        raise HTTPException(
            content=ExceptionContent(error=ApiError.OBJECT_NOT_FOUND, message=message)
        )


def gen_random_id(length: int = 20) -> str:
    """Generate a random base62 string (a-zA-Z0-9) of specified length. The max possible combinations is 62^length.
    Kucoin max 40, bingx max 40"""
    charset = string.ascii_letters + string.digits
    return "".join(random.choice(charset) for _ in range(length))


@typing_extensions.deprecated(
    "The `is_equal` method is deprecated; use `math.is_close` instead.", category=None
)
def is_equal(
    a: Union[float, Decimal],
    b: Union[float, Decimal],
    rel_tol: float = 1e-9,
    abs_tol: float = 0.0,
) -> bool:
    """
    Compare two Decimal numbers for approximate equality.
    """
    warnings.warn(
        "The `is_equal` method is deprecated; use `math.is_close` instead.",
        category=CrypticornDeprecatedSince25,
    )
    if not isinstance(a, Decimal):
        a = Decimal(str(a))
    if not isinstance(b, Decimal):
        b = Decimal(str(b))

    # Convert tolerances to Decimal
    return Decimal(abs(a - b)) <= max(
        Decimal(str(rel_tol)) * max(abs(a), abs(b)), Decimal(str(abs_tol))
    )


def optional_import(module_name: str, extra_name: str) -> Any:
    """
    Tries to import a module. Raises `ImportError` if not found with a message to install the extra dependency.
    """
    try:
        return __import__(module_name)
    except ImportError as e:
        raise ImportError(
            f"Optional dependency '{module_name}' is required for this feature. "
            f"Install it with: pip install crypticorn[{extra_name}]"
        ) from e


def datetime_to_timestamp(v: Any):
    """Converts a datetime to a timestamp.
    Can be used as a pydantic validator.
    >>> from pydantic import BeforeValidator, BaseModel
    >>> class MyModel(BaseModel):
    ...     timestamp: Annotated[int, BeforeValidator(datetime_to_timestamp)]
    """
    if isinstance(v, list):
        return [
            int(item.timestamp()) if isinstance(item, datetime) else item for item in v
        ]
    elif isinstance(v, datetime):
        return int(v.timestamp())
    return v



# coding: utf-8

"""
DEX AI API

API for DEX AI

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, ClassVar, Dict, List, Optional, Set, Union

from pydantic import BaseModel, ConfigDict, Field, StrictFloat, StrictInt, StrictStr
from typing_extensions import Self

from crypticorn.dex.client.models.signal_volume import SignalVolume
from crypticorn.dex.client.models.token_data import TokenData


class SignalWithToken(BaseModel):
    """
    Trading signal enriched with comprehensive token metadata and information.
    """  # noqa: E501

    ca: StrictStr = Field(description="The contract address of the token")
    name: StrictStr = Field(description="The name of the token")
    type: StrictStr = Field(description="The type of the signal")
    risk: Optional[StrictStr] = None
    performance: StrictStr = Field(
        description="The performance in percent as a string. e.g. +100%"
    )
    price: StrictStr = Field(description="The price of the signal")
    volume: SignalVolume = Field(description="The volume of the signal")
    liquidity: StrictStr = Field(description="The liquidity of the signal")
    mcap: StrictStr = Field(description="The market cap of the signal")
    called_at: StrictInt = Field(description="The timestamp of the call")
    updated_at: StrictInt = Field(description="The timestamp of the last update")
    tg_id: Optional[StrictStr]
    data: TokenData = Field(description="The token info")
    performance_float: Union[StrictFloat, StrictInt] = Field(
        description="The performance in float. e.g. +100% -> 2.0"
    )
    __properties: ClassVar[List[str]] = [
        "ca",
        "name",
        "type",
        "risk",
        "performance",
        "price",
        "volume",
        "liquidity",
        "mcap",
        "called_at",
        "updated_at",
        "tg_id",
        "data",
        "performance_float",
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of SignalWithToken from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        * OpenAPI `readOnly` fields are excluded.
        """
        excluded_fields: Set[str] = set(
            [
                "performance_float",
            ]
        )

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of volume
        if self.volume:
            _dict["volume"] = self.volume.to_dict()
        # override the default output from pydantic by calling `to_dict()` of data
        if self.data:
            _dict["data"] = self.data.to_dict()
        # set to None if risk (nullable) is None
        # and model_fields_set contains the field
        if self.risk is None and "risk" in self.model_fields_set:
            _dict["risk"] = None

        # set to None if tg_id (nullable) is None
        # and model_fields_set contains the field
        if self.tg_id is None and "tg_id" in self.model_fields_set:
            _dict["tg_id"] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of SignalWithToken from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "ca": obj.get("ca"),
                "name": obj.get("name"),
                "type": obj.get("type"),
                "risk": obj.get("risk"),
                "performance": obj.get("performance"),
                "price": obj.get("price"),
                "volume": (
                    SignalVolume.from_dict(obj["volume"])
                    if obj.get("volume") is not None
                    else None
                ),
                "liquidity": obj.get("liquidity"),
                "mcap": obj.get("mcap"),
                "called_at": obj.get("called_at"),
                "updated_at": obj.get("updated_at"),
                "tg_id": obj.get("tg_id"),
                "data": (
                    TokenData.from_dict(obj["data"])
                    if obj.get("data") is not None
                    else None
                ),
                "performance_float": obj.get("performance_float"),
            }
        )
        return _obj



# coding: utf-8

"""
Payment API

API for accepting payments and storing subscriptions

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, ClassVar, Dict, List, Optional, Set, Union

from pydantic import BaseModel, ConfigDict, Field, StrictFloat, StrictInt, StrictStr
from typing_extensions import Self

from crypticorn.pay.client.models.provider import Provider


class Invoice(BaseModel):
    """
    Model for reading an invoice
    """  # noqa: E501

    created_at: StrictInt = Field(description="Timestamp of creation")
    updated_at: StrictInt = Field(description="Timestamp of last update")
    id: StrictStr = Field(description="Unique identifier for the resource")
    product_id: StrictStr = Field(description="The ID of the product")
    coupon_id: Optional[StrictStr] = None
    provider: Provider = Field(description="The provider the invoice is created with")
    address: Optional[StrictStr] = None
    oob: Optional[StrictStr] = None
    user_id: StrictStr = Field(description="The ID of the user")
    usd_amount: Union[StrictFloat, StrictInt] = Field(
        description="The amount of the invoice in USD"
    )
    url: Optional[StrictStr] = None
    __properties: ClassVar[List[str]] = [
        "created_at",
        "updated_at",
        "id",
        "product_id",
        "coupon_id",
        "provider",
        "address",
        "oob",
        "user_id",
        "usd_amount",
        "url",
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of Invoice from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # set to None if coupon_id (nullable) is None
        # and model_fields_set contains the field
        if self.coupon_id is None and "coupon_id" in self.model_fields_set:
            _dict["coupon_id"] = None

        # set to None if address (nullable) is None
        # and model_fields_set contains the field
        if self.address is None and "address" in self.model_fields_set:
            _dict["address"] = None

        # set to None if oob (nullable) is None
        # and model_fields_set contains the field
        if self.oob is None and "oob" in self.model_fields_set:
            _dict["oob"] = None

        # set to None if url (nullable) is None
        # and model_fields_set contains the field
        if self.url is None and "url" in self.model_fields_set:
            _dict["url"] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of Invoice from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "created_at": obj.get("created_at"),
                "updated_at": obj.get("updated_at"),
                "id": obj.get("id"),
                "product_id": obj.get("product_id"),
                "coupon_id": obj.get("coupon_id"),
                "provider": obj.get("provider"),
                "address": obj.get("address"),
                "oob": obj.get("oob"),
                "user_id": obj.get("user_id"),
                "usd_amount": obj.get("usd_amount"),
                "url": obj.get("url"),
            }
        )
        return _obj
